/**********************************************************************
Run a servo off a clock.
Print out data concurrently.
Get position in radians
Estimate velocity in radians/second using differencing and low pass filter
Use velocity in control
/**********************************************************************/

#include "Pins.h"
#include "My_encoders.h"

/**********************************************************************/

#define MAX_ULONG 0xFFFFFFFF

#define SERVO_INTERVAL 2000 // microseconds

// start data collection after N milliseconds from go 
#define START_COLLECT  5000 // milliseconds
#define RAMP_DURATION 1000 // milliseconds
#define DEBUG_PRINT_INTERVAL 500 // milliseconds, 

#define MAX_COMMAND 255

// at x/sample should take 1s to get to 1.0
#define RAMP_INCREMENT (SERVO_INTERVAL*1e-6)

// #define PI 3.141592653589793
#define ENCODER_TO_RADIANS 0.004027682889218

/**********************************************************************/

bool go = false;  /* motor enable */

// safety ramp to have smooth startup and shutdown of motor
float ramp = 0;

// Encoder readings: these can be positive or negative.
long left_angle = 0;
long right_angle = 0;

// From Matlab butter( 2, 0.2 )
float a_filter[3] = {1.000000000000000, -1.142980502539901, 0.412801598096189};
float b_filter[3] = {0.067455273889072, 0.134910547778144, 0.067455273889072};

// velocity filter stuff
float left_diff[10];
float right_diff[10];
float past_left_angle_radians;
float past_right_angle_radians;
float past_left_velocity[10];
float past_right_velocity[10]
;
/**********************************************************************/
/**********************************************************************/

void motor_init()
{
  pinMode(AIN1, OUTPUT);
  pinMode(BIN1, OUTPUT);
  pinMode(PWMA_LEFT, OUTPUT);
  pinMode(PWMB_RIGHT, OUTPUT);
  digitalWrite(AIN1, HIGH);
  digitalWrite(BIN1, LOW);
  analogWrite(PWMA_LEFT, 0);
  analogWrite(PWMB_RIGHT, 0);
  pinMode(STBY_PIN, OUTPUT);
  digitalWrite(STBY_PIN, HIGH);
}

void motor_stop()
{
  digitalWrite(AIN1, HIGH);
  digitalWrite(BIN1, LOW);
  analogWrite(PWMA_LEFT, 0);
  analogWrite(PWMB_RIGHT, 0);
}

void motor_left_command( int speed )
{
    if ( speed >= 0 )
    {
      digitalWrite( AIN1, 1 );
      analogWrite( PWMA_LEFT, speed );
    }
  else
    {
      digitalWrite( AIN1, 0 );
      analogWrite( PWMA_LEFT, -speed );
    }
}

// reverses the sign of "speed"
void motor_right_command( int speed )
{
  if ( speed >= 0 )
    {
      digitalWrite( BIN1, 1 );
      analogWrite( PWMB_RIGHT, speed );
    }
  else
    {
      digitalWrite( BIN1, 0 );
      analogWrite( PWMB_RIGHT, -speed );
    }
}

/**********************************************************************/

void setup()
{

  Serial.begin( 1000000 );
  analogReference(INTERNAL1V1); // was INTERNAL on 328P, voltage
  while( !Serial );
  Serial.println( "servo2 version 1" );
  delay(1000); // Delay to make sure above message prints out.

  motor_init();
  Serial.println( "motor_init done." );
  delay(1000); // Delay to make sure above message prints out.

  Encoder_init( &left_angle, &right_angle );
  Serial.println( "Initialized encoders" );
  Serial.println( "Wheels should be off the ground."  );
  Serial.println( "Type g <return> to run test, s <return> to stop."  );
  Serial.println( "Typing window is at the top of the Arduino serial monitor window." );
  Serial.println( "Type into the main window of a Putty serial monitor window." );

  go = false; // start turned off
}

/******************************************************************/

// Take user input
void ProcessCommand()
{

  if ( Serial.available() <= 0 )
    return;

  int c = Serial.read();
  switch (c)
    {
      case 'S': case 's':
        Serial.println( "Stop!" );
        go = false;
        break;
      case 'G': case 'g':
      Serial.println( "Go!" );
        go = true;
        break;
      default:
        break;
    }
}

/**********************************************************************/
// note that int arguments have to be less than +/-32000

void run_servo( float angle_gain, float angle_integral_gain, 
                float angle_velocity_gain,
                float angle_start, int delay, float angle_goal,
                unsigned long run_time_limit, int collect_data )
{
  // How long do we run for?
  // add extra time for transient period, startup ramp happens during this time
  run_time_limit += START_COLLECT;
  // add extra time for ramp down
  unsigned long real_run_time_limit = run_time_limit + RAMP_DURATION;

  // Clocks and time management
  // used to keep track of time intervals in servo
  unsigned long last_micros = micros();
  unsigned long last_millis = millis();
  int servo_time = 0; // Servo clock: needs to be signed, can be int
  // how long have we been running? Safety feature
  unsigned long run_time = 0; // milliseconds
  // Clocks for printing and other stuff
  unsigned long debug_print_time = 0; // milliseconds

  // Keep track of late control cycles. 
  unsigned long servo_late = 0; // How many times am I late?
  unsigned long max_servo_late = 0; // What was the worst one?

  // state
  int started = 0;

  // Where do we start?
  float angle_desired = angle_start;
  // When do we move
  unsigned long move_time = delay + START_COLLECT;
  // Safe startup control
  float safe_angle_gain = 300.0;  
  float safe_angle_integral_gain = 1.0;
  float safe_angle_velocity_gain = 10.0;  
  float safe_angle_error_integral_left = 0;
  float safe_angle_error_integral_right = 0;
  // actual control
  float angle_error_integral_left = 0;
  float angle_error_integral_right = 0;

  // keep track of voltage
  int voltage_raw = analogRead(VOL_MEASURE_PIN); //Read voltage value
  double voltage = (voltage_raw*1.1/1024)*((10+1.5)/1.5);
  if ( collect_data )
    {
      Serial.print( "Current voltage " );
      Serial.print( voltage );
      Serial.print( " " );
      Serial.println( voltage_raw );
    }

  // print out the angle_gain
  if ( collect_data )
    {
      Serial.print( "Gains " );
      Serial.print( angle_gain );
      Serial.print( " " );
      Serial.print( angle_integral_gain );
      Serial.print( " " );
      Serial.println( angle_velocity_gain );
    }

  // servo loop
  for( ; ; )
    {

/*****************************************
Timekeeping part */

      unsigned long current_micros = micros();
      if ( current_micros > last_micros )
        servo_time += current_micros - last_micros;
      else // rollover
        servo_time += (MAX_ULONG - last_micros) + current_micros;
      last_micros = current_micros;

      // It isn't time yet ...
      if ( servo_time < SERVO_INTERVAL )
        continue;

      // It is time! Reset the servo clock.
      servo_time -= SERVO_INTERVAL;

      // Keep track of maximum lateness
      if ( max_servo_late < servo_time )
        max_servo_late = servo_time;

      // Let's have some slop in counting late cycles.
      if ( servo_time > 50 )
        servo_late += 1;
    
      // handle the millisecond clocks
      unsigned long current_millis = millis();
      int millis_increment;
      if ( current_millis > last_millis )
        millis_increment = current_millis - last_millis;
      else // rollover
        millis_increment = (MAX_ULONG - last_millis) + current_millis;
      last_millis = current_millis;
  
      run_time += millis_increment;
      debug_print_time += millis_increment;

/*****************************************
Administrative part */

      // Turn off after a while to keep from running forever
      if ( ( run_time > real_run_time_limit ) || ( !go ) )
        {
          motor_stop();
          ramp = 0;
          Serial.println( "Motor stopped." );
          Serial.println( "Wheels should be off the ground."  );
          Serial.println( "Type g <return> to run test, s <return> to stop." );
          return;
        }

/*****************************************
State estimation part */

      // Read the sensors
      Read_encoders( &left_angle, &right_angle );

      float left_angle_radians = ENCODER_TO_RADIANS*left_angle;
      // float right_angle_radians = ENCODER_TO_RADIANS*right_angle;

//     a(1)*y(n) = b(1)*x(n) + b(2)*x(n-1) + ... + b(nb+1)*x(n-nb)
//                 - a(2)*y(n-1) - ... - a(na+1)*y(n-na)
      left_diff[2] = left_diff[1];
      left_diff[1] = left_diff[0];
      // this is radians/sec - diff/Ts
      left_diff[0] = 500*(left_angle_radians - past_left_angle_radians);
      past_left_angle_radians = left_angle_radians;
      float left_velocity = b_filter[0]*left_diff[0]
                      + b_filter[1]*left_diff[1]
                      + b_filter[2]*left_diff[2]
		      - a_filter[1]*past_left_velocity[0]
		      - a_filter[2]*past_left_velocity[1];
      past_left_velocity[1] = past_left_velocity[0];
      past_left_velocity[0] = left_velocity;
/*
      right_diff[2] = right_diff[1];
      right_diff[1] = right_diff[0];
      // this is radians/sec - diff/Ts
      right_diff[0] = 500*(right_angle_radians - past_right_angle_radians);
      past_right_angle_radians = right_angle_radians;
      float right_velocity = b_filter[0]*right_diff[0]
                      + b_filter[1]*right_diff[1]
                      + b_filter[2]*right_diff[2]
		      - a_filter[1]*past_right_velocity[0]
		      - a_filter[2]*past_right_velocity[1];
      past_right_velocity[1] = past_right_velocity[0];
      past_right_velocity[0] = right_velocity;
      */

/*****************************************
Servo part */

      if ( run_time >= move_time )
        angle_desired = angle_goal;

      float left_error = left_angle_radians - angle_desired;
      // float right_error = right_angle_radians - angle_desired;

      float left_commandf = 0;
      // float right_commandf = 0;
      
      if ( run_time < START_COLLECT )
        { // use safe gain during startup
          if ( abs( left_error ) < 0.1 )
            safe_angle_error_integral_left
              -= safe_angle_integral_gain*left_error;
          left_commandf = - safe_angle_gain*left_error
	                  - safe_angle_velocity_gain*left_velocity
			  + safe_angle_error_integral_left;
      /*
          if ( abs( right_error ) < 0.1 )
            safe_angle_error_integral_right
              -= safe_angle_integral_gain*right_error;
          right_commandf = -safe_angle_gain*right_error
            + safe_angle_error_integral_right;
      */
      	}
      else
        {
          if ( abs( left_error ) < 0.1 )
            angle_error_integral_left
              -= angle_integral_gain*left_error;
    	  left_commandf = - angle_gain*left_error
	                  - angle_velocity_gain*left_velocity
    	                  + angle_error_integral_left;
	  /*
          if ( abs( right_error ) < 0.1 )
            angle_error_integral_right
              -= angle_integral_gain*right_error;
      
          right_commandf = -angle_gain*right_error + angle_error_integral_right;
	  */
       	}

      // ramp command up at beginning
      if ( ( run_time < START_COLLECT )
           && ( ramp < (1.0 - RAMP_INCREMENT) ) )
        ramp += RAMP_INCREMENT;

      // ramp command down at end
      if ( run_time > run_time_limit )
        ramp *= 0.95; // gradually turn motor off

      int left_command = round( ramp*left_commandf );
      // int right_command = round( ramp*right_commandf );
      int right_command = 0;

      if ( left_command > MAX_COMMAND )
        left_command = MAX_COMMAND;
      if ( left_command < -MAX_COMMAND )
        left_command = -MAX_COMMAND;
/*
      if ( right_command > MAX_COMMAND )
        right_command = MAX_COMMAND;
      if ( right_command < -MAX_COMMAND )
        right_command = -MAX_COMMAND;
	*/

      motor_left_command( left_command );
      motor_right_command( right_command );

/*****************************************
Data printing part */

      // print out debugging info before data collection
      if ( ( debug_print_time > DEBUG_PRINT_INTERVAL )
           && ( ( run_time < START_COLLECT )
	        || ( run_time > run_time_limit ) ) )
        {
          Serial.print( servo_late );
          Serial.print( " " );
          Serial.print( max_servo_late );
          Serial.print( " " );
          Serial.print( left_error );
          Serial.print( " " );
	  /*
          Serial.print( right_error );
          Serial.print( " " );
	  */
          Serial.println( ramp );
          ProcessCommand();
          debug_print_time -= DEBUG_PRINT_INTERVAL;
        }

      // print out data
      if ( ( run_time >= START_COLLECT ) && ( run_time <= run_time_limit )
           && collect_data )
        {
          if ( !started )
            {
              servo_late = 0;
              max_servo_late = 0;
              started = 1;
              Serial.println( "Data" );
            }
          Serial.print( current_micros );
          Serial.print( " " );
          Serial.print( left_angle );
          Serial.print( " " );
          // Serial.print( right_angle );
          // Serial.print( " " );
          Serial.print( left_command );
          Serial.print( " " );
          // Serial.print( right_command );
          //Serial.print( " " );
          Serial.println( left_velocity);
          // Serial.print( " " );
          // Serial.println( right_velocity);
	  // Serial.print( " " );
          // Serial.print( left_angle_radians);
          // Serial.print( " " );        
          // Serial.println( left_error ); 
        }
    }
}

/**********************************************************************/

void loop()
{
  if ( !go )
    {
      motor_stop();
      delay(500);
      ProcessCommand();
      return;
    }

  // run an experiment
  run_servo( 60.0, 0.0, 0.0, 0, 100, PI, 2000, 1 );
  // get the wheels back to zero
  delay( 1000 );
  run_servo( 300.0, 1.0, 10.0, 0, 100, 0, 1000, 0 );
  delay( 1000 );

  run_servo( 70.0, 0.0, 1.0, 0, 100, PI, 2000, 1 );
  // get the wheels back to zero
  delay( 1000 );
  run_servo( 300.0, 1.0, 10.0, 0, 100, 0, 1000, 0 );
  delay( 1000 );

  run_servo( 80, 0.0, 2.0, 0, 100, PI, 2000, 1 );
  // get the wheels back to zero
  delay( 1000 );
  run_servo( 300.0, 1.0, 10.0, 0, 100, 0, 1000, 0 );
  delay( 1000 );

  Serial.println( "Stop!" );
  go = false;
}
  
/**********************************************************************/
